// SilberLand MUDlib
//
// /p/daemon/channeld.c -- Channel daemon
//
// $Date: 1997/06/25 10:44:05 $
/* $Revision: 1.9 $
 * $Log: channeld.c,v $
 * Revision 1.9  1997/06/25 10:44:05  Woody
 * Message() Aufruf um MSGFLAG_ erweitert, nun duerfen auch
 * Clones von trusted objects senden
 *
 * Revision 1.7  1997/04/24 23:48:06  Woody
 * add_hist() und dump_hist() sind jetzt private (security fix)
 *
 * Revision 1.6  1997/04/14 14:55:20  Woody
 * verwendet auch bei InterMud get_name (obj->name())
 *
 * Revision 1.5  1997/04/08 00:29:54  Woody
 * Schoenere Formatierung bei Channel (g)emotes
 *
 * Revision 1.4  1997/03/16 02:24:20  Woody
 * prepared for revision control
 *
 * Revision 1.3  09.10.1996  00:00:00  Woody
 * P_PERM_STRING wird jetzt ausgewertet
 *
 * Revision 1.2  05.03.1996  00:00:00  Hadra
 * Spieler werden bei -!! keine unsichtbaren Magier mehr angezeigt
 *
 * Revision 1.1  xx.xx.xxxx  xx:xx:xx  Mupfel
 * Einfuehrung der Sprachveraenderung auf dem Unsinnskanal
 *
 * Revision 1.0  xx.xx.xxxx  xx:xx:xx  Unbekannt
 * Alte Logs aufgeraeumt
 *
 */

/*
 * CHANNELD.C -- Der Daemon fuer die Chatkanaele im Mud
 * AnderLand Mudlib
 *
 * Author:
 *   Hate@MorgenGrauen
 *   Weiterentwicklung im AL siehe Revisionsgeschichte
 * Zweck:
 *   Der Channeldaemon regelt die Benutzung der mudinternen Kaenele und die
 *   muduebergreifende Kommunikation auf den D- und I-Kanaele
 */

#define NEED_PROTOTYPES

#include <thing/properties.h>
#include <daemon.h>
#include <properties.h>
#include <wizlevels.h>
#include <udp.h>
#include <defines.h>
#include <terminal.h>
#include <ansi.h>

#define DEBUG(x)     if(find_player("querolin")) tell_object(find_player("querolin"), sprintf("-> %O\n", x))

#define SAR (["MG ":"MorgenGrauen ", "AL ":"AnderLand ", \
              " MG":" MorgenGrauen", " AL":" AnderLand" ])

#define USER         0
#define LETTER       1
#define MRECV        2
#define MSEND        3
#define PRECV        4
#define PSEND        5
#define FLAGS        6
#define INFO         7
#define HIST	     8
#define MAX_ENTRIES  9	// Erhoehen, wenn was Neues dazu kommt!

#define POSSIBLE_CHANNELS        "Es stehen folgende Ebenen zu Verfuegung:\n"
#define REGISTERED_CHANNELS      "Auf folgenden Ebenen wird gelauscht:\n"
#define LIST_FMT                 "[%s] %-'.'15s (%2d) %s"

#define MESG(ch, u, msg)         (previous_object() != ME?\
                                  (int)tell_object(u, sprintf(msg, ch)):0)
#define REGISTERED               "%s wurde schon betreten.\n"
#define NOT_REGISTERED           "%s wurde nicht betreten.\n"
#define NO_SUCH_CHANNEL          "An die Ebene %s darfst Du nicht ran.\n"
#define NO_SEND                  "Auf Ebene %s darf nicht gesendet werden.\n"

#define ENTER                    "Du betrittst die Ebene %s.\n"
#define LEAVE                    "Du verlaesst die Ebene %s.\n"
#define LISTENERS                "Folgene Spieler hoeren auf %s:\n"

#define SEND(ch, u, msg)         funcall(#'SendToChannel, ch, u, SAY, msg)
#define NEW_CHANNEL              "New channel:"
#define DAEMON_BOOT              "Reboot initiated by %O..."
#define DAEMON_SHUTDOWN          "Shutdown initiated by %O..."
#define CHANNEL_INFO             "%d channels with %d users active."
#define REGISTER_USER            "Registered %s on %s."
#define REMOVE_USER              "Removed %s from %s."

#define WIZ                      1000
#define ISWIZLEV(l)              (l >= LEARNER_LVL*WIZ)
#define TRUSTED                  150000
#define CHECK_SEND(l, ch)        check_level(1, l, ch)
#define CHECK_RECV(l, ch)        check_level(0, l, ch)

#define GHOSTS                   "/p/daemon/german.hosts"
#define HIST_SIZE		 15

private mapping channels;
static private mixed trusted;

/*
 * Hilfsfunktionen
 */

private string strip_space(string str)
{ return implode(efun::explode(str, " ")-({""}), " "); }

private int is_invis_wiz(object ob)
{
  return (IS_WIZARD(ob) && ob->QueryProp(P_INVIS));
}

/*
 * Ende Hilfsfunktionen
 */

private int create_channel(string cmd)
{
  mixed cmds;
  if(!stringp(cmd) || !strlen(cmd) || cmd[0] == '#') return -1;
  cmds = map_array(explode(cmd, ":"), #'strip_space);
  if(pointerp(cmds) && sizeof(cmds) &&
     !member(channels, cmds[0]))
  {
    switch(sizeof(cmds)-1)
    {
    case HIST  : channels[cmds[0], HIST]   = 0;
    case INFO  : channels[cmds[0], INFO]   = cmds[7];
    case FLAGS : channels[cmds[0], FLAGS]  = to_int(cmds[6]);
    case PSEND : channels[cmds[0], PSEND]  = to_int(cmds[5]);
    case PRECV : channels[cmds[0], PRECV]  = to_int(cmds[4]);
    case MSEND : channels[cmds[0], MSEND]  = to_int(cmds[3]);
    case MRECV : channels[cmds[0], MRECV]  = to_int(cmds[2]);
    case LETTER: channels[cmds[0], LETTER] = cmds[1];
    case USER  : channels[cmds[0], USER]   = ({});
		 channels[cmds[0], HIST]   = 0; // Auf jeden Fall initialisieren
		 break;
    }
    SEND("Channel", ME, NEW_CHANNEL+" '"+implode(cmds, ":")+"'");
  }
}

void create()
{
  mixed init_data, u;
  object pl;
  int i;
  if(is_clone(ME)) destruct(ME);
  channels = allocate_mapping(0,MAX_ENTRIES);
  channels += (["Channel":({});"C";20; -1; -1; -1; 0; "Channel Information"; 0]);
  trusted  = ([__FILE__[0..<3]: ""; ""]);
  if(init_data = read_file(file_name(ME)+".trusted"))
  {
    i = sizeof(init_data = map_array(explode(init_data, "\n")-({}),
	       lambda(({'s}), ({#'map_array, ({#'explode, 's, ":"}),
		                             #'strip_space}))));
    while(i--)
    {
      if(sizeof(init_data[i]) && init_data[i][0][0] != "#")
        trusted+=([init_data[i][0]:(sizeof(init_data[i])>1)?init_data[i][1]:"";
		                   (sizeof(init_data[i])>2)?init_data[i][2]:""
                 ]);
    }
  }
  if(init_data = read_file(file_name(ME)+".init"))
    map_array(explode(init_data, "\n"), #'create_channel, ME);
  u = map_array(users(), #'RegisterUser);
  SEND("Channel", ME, sprintf(DAEMON_BOOT, previous_object()));
  SEND("Channel", ME, sprintf(CHANNEL_INFO,m_sizeof(channels),sizeof(u-({0}))));
}

int remove()
{
  mixed user; user = ({});
  SEND("Channel", ME, sprintf(DAEMON_SHUTDOWN, previous_object()));
  map_array(m_values(channels), #'map_array, 
            lambda(({'u, 'us}), ({#'?, ({#'==, ({#'member, 'us, 'u}), -1}),
                                      ({#'+=, 'us, ({#'({, 'u})})})), &user);
  map_array(user, #'RemoveUser);
  destruct(ME);
  return 1;
}

int query_prevent_shadow() { return 1; }

varargs private int secure(object user, string channel)
{
  int i;
  string fname;
  if(!objectp(user)) return 0;
  if(member(trusted, fname = explode(file_name(user),"#")[0]) &&
      user == previous_object())
    if(!channel) return TRUSTED;
    else if(!strlen(trusted[fname]) ||
            member(trusted[fname], channel[0]) != -1)
           return TRUSTED;
  if(!query_once_interactive(user)) return 0;
  return query_wiz_level(user)*WIZ + (user->QueryProp(P_LEVEL) % 1000);
}

private int check_level(int send, int level, string ch)
{
  int ok, l;
  if(level == TRUSTED) return 1;

  // wizlevel restrictions
  l=channels[ch, send?MSEND:MRECV];
  ok = (l != -1) && (l <= level/WIZ);

  // playerlevel restrictions
  l=channels[ch, send?PSEND:PRECV];
  ok = ok || ((l != -1) && (l <= level%WIZ));

  if(ok && (channels[ch, FLAGS] & LOCK))
    if(ISWIZLEV(level) && channels[ch, MSEND] <= level/WIZ) return 1;
    else return 0;
  return ok;
}

varargs private string get_name(mixed pl, int level, int fall)
{
  string file, tmp;
  if (!objectp(pl)) {
    if (pointerp(pl))
      return pl[ISWIZLEV(level)];
    else
      return "<Unbekannt>";
  }
  if(member(trusted, file = file_name(pl))) 
    if(tmp = file->name()) return capitalize(tmp); 
    else return trusted[file, 1];
  if(pl->QueryProp(P_INVIS))
    return ISWIZLEV(level) ? ("("+CAP(getuid(pl))+")") : "Jemand";
  else
    return pl->QueryProp(P_FROG)?("Frosch "+CAP(getuid(pl))):pl->name(fall);
}

private int is_channel(string channel, string ch)
{
  return ch == channels[channel, LETTER];
}

private mixed get_channel(string channel)
{
  int i;
  mixed tmp,ch; ch = ({});
  tmp=m_indices(channels);
  if(!strlen(channel)) return tmp;
  channel = implode(regexplode(channel, "[.?*\^|/\\\\]")
          - ({".", "?", "*", "^", "|", "/", "\\"}), "");
  if(!strlen(channel)) return ({});
  for (i=sizeof(tmp)-1;i>=0;i--)
    if (channels[tmp[i],LETTER]==channel)
	  ch+=({tmp[i]});
  if(!sizeof(ch)) return regexp(m_indices(channels), "^"+channel+".*");
  return ch;
}

int RegisterUser(object user)
{
  mixed plchannels;
  if(pointerp(plchannels = user->QueryProp(P_CHANNELS)) &&
     sizeof(plchannels))
  {
    map_array(plchannels, #'RegisterChannel, user);
    SEND("Channel",ME,sprintf(REGISTER_USER, get_name(user, TRUSTED), 
                                             implode(plchannels, ", ")));
    return 1;
  }
}

private void remove_user(string channel, object user, mixed plchannels)
{
  if(member(channels[channel], user) != -1) 
  {
    plchannels += ({ channel });
    channels[channel] -= ({0, user});
  }
}

int RemoveUser(object user)
{
  mixed plchannels; plchannels = ({});
  if(!secure(user)) return -1;
  map_mapping(channels, #'remove_user, user, &plchannels);
  if(sizeof(plchannels))
   SEND("Channel", ME, sprintf(REMOVE_USER, get_name(user, TRUSTED), 
                                            implode(plchannels, ", ")));
  user->SetProp(P_CHANNELS, plchannels);
}

int RegisterChannel(string channel, object user)
{
  mixed ch;
  int i, j, level;
  if(!(j = i = sizeof(ch = get_channel(channel))))
    return MESG(channel, user, NO_SUCH_CHANNEL);
  while(j--)
    if(!(level = secure(user, channels[ch[j], LETTER]))) { ch[j] = ({}); i--; }
  if(!sizeof(ch)) return MESG(channel, user, NO_SUCH_CHANNEL);
  while(i--)
    if(CHECK_RECV(level, ch[i])) 
      if(level == TRUSTED || member(channels[ch[i]], user) == -1)
      {
        mixed tmp;
        channels[ch[i]] = (channels[ch[i]] - ({0})) + ({ user });
        if(member(pointerp(tmp = user->QueryProp(P_CHANNELS))?tmp:(tmp = ({})),
      ch[i]) == -1)
    user->SetProp(P_CHANNELS, tmp + ({ ch[i] }));
        MESG(ch[i], user, ENTER);
      }
      else MESG(ch[i], user, REGISTERED);
    else MESG(ch[i], user, NO_SUCH_CHANNEL);
}

int RemoveChannel(string channel, object user)
{
  mixed ch;
  int i, j, level;
  if(!(j = i = sizeof(ch = get_channel(channel))))
    return MESG(channel, user, NO_SUCH_CHANNEL);
  while(j--)
    if(!(level = secure(user, channels[ch[j], LETTER]))) { ch[j] = ({}); i--; }
  if(!sizeof(ch)) return MESG(channel, user, NO_SUCH_CHANNEL);
  while(i--)
    if(CHECK_RECV(level, ch[i]))  
      if(level = TRUSTED || member(channels[ch[i]], user) != -1)
      {
        channels[ch[i]] -= ({0, user });
  user->SetProp(P_CHANNELS, user->QueryProp(P_CHANNELS) - ({ ch[i] }));
        MESG(ch[i], user, LEAVE);
      }
      else MESG(ch[i], user, NOT_REGISTERED);
    else MESG(ch[i], user, NO_SUCH_CHANNEL);
} 

private void print_channel(string channel, mixed luser, string letter,
                           int precv, int psend, int mrecv, int msend,
                           int flags, string info, mixed hist, object user,
                           int level, mixed message)
{
  luser -= ({0});
  if(!CHECK_RECV(level, channel)) return;
  message += ({ sprintf(LIST_FMT, letter, channel, 
  (IS_WIZARD(user)?sizeof(luser):(sizeof(luser)-(sizeof(filter_array(luser, #'is_invis_wiz))))),
                        stringp(info)?info:"") });
}

varargs int ListChannels(object user)
{
  int level;
  mixed message; message = ({});
  if(!objectp(user))
  {
    if(!(level = secure(user = PL))) return -1;
    tell_object(user, POSSIBLE_CHANNELS);
    walk_mapping(channels, #'print_channel, user, level, &message);
  }
  else
  {
    if(!(level = secure(user))) return -1;
    tell_object(user, REGISTERED_CHANNELS);
    walk_mapping(filter_mapping(channels, 
                                lambda(({ 'c }), 
                                       ({#'>, ({#'member, 
                                                ({#'[, channels, 'c }), user}),
                                              -1 }) )),
                 #'print_channel, user, level, &message);
  }
  tell_object(user, implode(sort_array(message, #'>), "\n")+"\n");
}

int ListUsers(string channel, object user)
{
  mixed ch;
  int i, j, level;
  if(!(j = i = sizeof(ch = get_channel(channel))))
    return MESG(channel, user, NO_SUCH_CHANNEL);
  while(j--)
    if(!(level = secure(user, channels[ch[j], LETTER]))) { ch[j] = ({}); i--; }
  if(!sizeof(ch)) return MESG(channel, user, NO_SUCH_CHANNEL);
  while(i--)
    if(CHECK_RECV(level, ch[i]))
      if(level == TRUSTED || member(channels[ch[i]], user) != -1)
        tell_object(user, sprintf(LISTENERS, ch[i])+break_string(implode(
		                  sort_array(
                          map_array(channels[ch[i]] -= ({0}), #'get_name, 
                                    level, WER)-({"Jemand"}), #'>), ", "), 78));
      else MESG(ch[i], user, NOT_REGISTERED);
    else MESG(ch[i], user, NO_SUCH_CHANNEL);
}

private varargs void tell_user(object receiver, string channel, 
                       string msg, mixed sender, int cmd, int flags)
{
  string pre, race;
  int fall, c;
  mixed res;
  // ::: mupfel start ***
  if (!IS_LEARNING(receiver))
  {
    race = receiver->QueryProp(P_RACE);
    if (channel=="Elfen" && race!="Elf")
      msg = "elfensingsangsingsanglalala";
    if (channel=="Hobbits" && race!="Hobbit")
      msg = "hobbitgefluestertuschel";
    if (channel=="Menschen" && race!="Mensch")
      msg = "laberrhabarberblafasel";
    if (channel=="Zwerge" && race!="Zwerg")
      msg = "zwergengemurmelmurmelmurmel";
    if (channel=="Brownies" && race!="Brownie")
      msg = "brownieblablablublaeblablu";
    if (channel=="Orks" && race!="Ork")
      msg = "Orkuggugguggorkgutuggugg";

  }
  // *** mupfel end ***
  switch(cmd)
  {
  case EMPTY : break;
  case GEMOTE: fall = WESSEN;
  case EMOTE : msg = sprintf("[%s:%s]", 
                             channel,
                             break_string(
                               get_name(sender, secure(receiver),fall)+" "+msg,
                               78, strlen(channel)+2)[strlen(channel)+2..<2]);
               pre = ""; break;
  case SAY   : pre = sprintf("[%s:%s] ", 
                             channel,
                             get_name(sender, secure(receiver),fall)); 
               break;
  }
  msg = cmd != SAY ? msg+"\n" : break_string(msg, 78, pre);
  // msg = cmd==EMPTY ? msg+"\n" : break_string(msg, 78, pre);
  //
  
  if(channel=="Unsinn") 
  for(c=strlen(msg)-1;c >= 0; c--)
  {
    if(msg[c]=='r') msg[c]='\'';
    if(msg[c]=='R') msg[c]='\'';
    if(msg[c]=='s') msg[c]='f';
    if(msg[c]=='S') msg[c]='F';
  }
  
#ifdef SEARCHANDREPLACE
  // *** woody: search-and-replace part ***
  {
    string s1, s2;
    object *ind;
    int i;

    ind = m_indices(SAR);
    s2 = msg;
    for (i=sizeof(SAR); i>0;)
    {
      i--;
      s2=msg;
      msg="";
      while (sscanf(s2, "%s"+ind[i]+"%s", s1, s2) == 2)
      {
        msg+=s1+SAR[ind[i]];
      }
      msg+=s2;
    }
  }
  // *** woody: end ***
#endif
  if (interactive(receiver)) {
  receiver->Set(P_COLOR_CHANNELD, SAVE, F_MODE_AS);
  if (receiver->QueryProp(P_TTY)=="ansi")
  {
   string s1; 
   switch(to_string(receiver->QueryProp(P_COLOR_CHANNELD))) {
   
     case "1": s1=ANSI_BLACK; break;
     case "2": s1=ANSI_RED; break;
     case "3": s1=ANSI_GREEN; break;
     case "4": s1=ANSI_BOLD+ANSI_YELLOW; break;
     case "5": s1=ANSI_BLUE; break;
     case "6": s1=ANSI_PURPLE; break;
     case "7": s1=ANSI_CYAN; break;
     case "8": s1=ANSI_WHITE; break;
     case "9": s1=ANSI_NORMAL+ANSI_BOLD; break;
     case "10": s1=ANSI_BOLD+ANSI_BLUE; break; //hellblau
     case "11": s1=ANSI_YELLOW; break;  //braun
     case "12": s1=ANSI_BOLD+ANSI_BLACK; break; //grau
     case "13": s1=ANSI_BOLD+ANSI_CYAN; break; //hellcyan
     case "14": s1=ANSI_BOLD+ANSI_GREEN; break; //hellgruen
     case "15": s1=ANSI_BOLD+ANSI_RED; break; //hellrot
     case "16": s1=ANSI_BOLD+ANSI_PURPLE; break; //helllila
     default: s1=ANSI_NORMAL;
   }
    msg=s1+msg+ANSI_NORMAL; 
   
  }
  }
  res=receiver->Message(msg, flags || MSGFLAG_CHANNEL);
}

private void send_intermud(string mud, mixed data, mapping msg, string *excl)
{
  if(member(data[HOST_COMMANDS], "channel") != -1 ||
     member(data[HOST_COMMANDS], "*") != -1)
  {
    if(excl && member(excl, mud) == -1) return;
    INETD->send_udp(data[HOST_NAME], msg);
  }
}

private void add_hist(string ch, string msg, object user, int cmd)
{
  if (!pointerp(channels[ch, HIST]))
    channels[ch, HIST] = allocate(HIST_SIZE);

  channels[ch, HIST] = ({ ({ time(),
			     msg,
			     ({ get_name(user, 0,
                                    cmd == GEMOTE ? WESSEN : WER),
				get_name(user, LEARNER_LVL*WIZ,
                                    cmd == GEMOTE ? WESSEN : WER)}),
			     cmd})
		       }) + channels[ch, HIST][0..HIST_SIZE-2];
}

private void dump_hist(string ch, object recv)
{
  mixed hist;
  int i;

  hist = channels[ch, HIST];
  if (!hist || !sizeof(hist -= ({ 0 }) ))
    tell_user(recv, ch, "Allein sphaerisches Rauschen erfuellt diese Ebene.",
              this_object(), SAY, MSGFLAG_CHIST);
  else {
    tell_user(recv, ch, "Was bisher geschah:",
              this_object(), SAY, MSGFLAG_CHIST);
    for (i=sizeof(hist)-1; i>=0; i--) {
      if (hist[i]) {
	string zeit;
    
	if (time()-hist[i][0] < 18*3600)
	  zeit = dtime(hist[i][0])[19..];
	else {
	  zeit = dtime(hist[i][0])[5..];
	  if (zeit[0] == ' ')
	    zeit = zeit[1..];
	}
	tell_user(recv, ch, hist[i][1] + " <"+zeit+">",
                  hist[i][2], hist[i][3], MSGFLAG_CHIST);
      }
    }
    tell_user(recv, ch, "----------------", this_object(), SAY);
  }
}

varargs int SendToChannel(string channel, object user, int cmd, string msg)
{
  mixed ch;
  int i, j, level;
  string fmt;

//  msg = implode(explode(implode(explode(msg,"\b"),""),ESC),"");
  if(!(j = i = sizeof(ch = get_channel(channel)))) 
    return MESG(channel, user, NO_SUCH_CHANNEL);
  while(j--)
    if(!(level = secure(user, channels[ch[j], LETTER]))) { ch[j] = ({}); i--; }
  if(!sizeof(ch)) return MESG(channel, user, NO_SUCH_CHANNEL);
  // woody
  // note: eigentlich sollte der Sender seinen Sprachfluch nicht sehen,
  //       aber fuer jeden Empfaenger eigens permutate_string aufzurufen
  //       halte ich fuer zu aufwendig, und ne globale Variable oder
  //       ne andere Hackerei wollt ich mir jetzt nicht antun...
  if (objectp(user) && objectp(user->QueryProp(P_PERM_STRING)))
    msg=call_other(user->QueryProp(P_PERM_STRING), "permutate_string", (msg))||msg;
  while(i--)
    if(CHECK_SEND(level, ch[i]))
      if(level == TRUSTED || member(channels[ch[i]], user) != -1)
      {
        channels[ch[i]] -= ({0});
        if(msg[<1]=='\n') msg = msg[0..<2];
        if((ch[i][0..4]=="Inter"||ch[i][0..1]=="D-") &&
	   user != find_object("/secure/udp/channel"))
        {
          mapping request; string *excl;
          request = ([REQUEST : "channel",
                      SENDER  : get_name(user),
                     "CHANNEL": lower_case(ch[i]),
                      DATA    : implode(explode(msg, "\n"), " ")]);
          if(cmd == GEMOTE || cmd == EMOTE)
          {
            request["EMOTE"] = 1;
            request["CMD"] = "emote";
          }
          /*
          if(ch[i][0..1] == "D-")
	  {
            if(file_size(GHOSTS) < 0) return MESG(ch[i], user, NO_SEND);
            excl = map_array(explode(read_file(GHOSTS), "\n"), #'lower_case);
          }
          */
          walk_mapping(INETD->query("hosts"), #'send_intermud, request, excl);
        }
	else
	{
          if(ch[i]=="Debug") 
          {
            int s;
            msg = implode(explode(msg, "::"), " ");
	    add_hist(ch[i], msg, user, cmd);
            for(s = sizeof(channels[ch[i]])-1; s >= 0; s--)
              tell_user(channels[ch[i]][s], ch[i], msg, user, cmd);
          }
          else
          {
            map_array(channels[ch[i]],#'tell_user,ch[i],msg,user,cmd);
	    add_hist(ch[i], msg, user, cmd);
          }
	}
      }
      else MESG(ch[i], user, NOT_REGISTERED);
    else MESG(ch[i], user, NO_SEND);
}

int ChannelHistory(string channel, object user)
{
  mixed ch;
  int i, j, level;

  if(!(j = i = sizeof(ch = get_channel(channel))))
    return MESG(channel, user, NO_SUCH_CHANNEL);
  while(j--)
    if(!(level = secure(user, channels[ch[j], LETTER]))) { ch[j] = ({}); i--; }
  if(!sizeof(ch)) return MESG(channel, user, NO_SUCH_CHANNEL);
  while(i--)
    if(CHECK_RECV(level, ch[i]))
      dump_hist(ch[i], user);
    else MESG(ch[i], user, NO_SUCH_CHANNEL);
}

