// SilberLand MUDlib
//
// std/food.c -- standard food object
//
// $Date: 2000/06/12 09:16:30 $
/* $Revision: 1.2 $
 * $Log: food.c,v $
 * Revision 1.2  2000/06/12 09:16:30  Samtpfote
 * essen, dass sich in Spielern oder Raeumen befindet, loest sich nun
 * nach 1-3 Resets von selber auf.
 *
 * Revision 1.1  2000/06/12 09:08:41  Samtpfote
 * Initial revision
 *
 * Revision 1.2  1997/10/07 11:45:11  Woody
 * Ueberarbeitet, nun wesentlich flexibler
 *
 * Revision 1.1  1997/03/13 02:53:57  Woody
 * prepared for revision control
 *
 * 20.10.1995	00:58	Mupfel	P_MAX_FOOD, P_MAX_ALCOHOL selber testen
 * 13.10.1995	19:54	Mupfel	/sys/food.h : F_HEAL statt "heal" etc.
 * 06.09.1995	04:00	Mupfel	erste Version
 */

// Properties:
// P_FOOD_INFO : mapping
// ([ F_HEAL:mixed,       Heilung, int (LP+SP) oder array ({LP, SP})
//    F_SOAK:int,         Saettigung
//    F_DRINK:int,        Trink-Saettigung
//    F_ALCOHOL:int,      Alkoholgehalt
//    F_POISON:int,       Giftgehalt
//    F_MSG:string,       Text beim Essen/Trinken fuer den Spieler
//    F_MSG_ROOM:string,  Text beim Essen/Trinken fuer den Raum
//                        (Spielername+" " wird automatisch vorangestellt)
//    F_VERB:string ])    Verb (wird bei der Fehlermeldung benoetigt);
//                        Default ist "essen" (Verb in Nennform!!)
//
// (F_HEAL und F_POISON wird NACH 'BeimEssen' gesetzt, F_SOAK und
// F_ALCOHOL davor, um abzutesten, ob der Spieler das noch essen kann)
//
// Funktionen:
//
// void BeimEssen();
//   Wird in der 'Essfunktion' aufgerufen. Das Nahrungsobjekt wird
//   anschliessend zerstoert, wenn die Funktion second_life() nicht
//   existiert oder 0 zurueckliefert.
//
// int second_life(object pl);
//   Wenn diese Funktion einen Wert ungleich 0 zurueckgibt, so hat
//   das Objekt ein "zweites Leben" und wird nicht destructed nach
//   der Konsumierung. Hier koennte man auch das Clonen leerer
//   Flaschen, Obstkerne, etc. implementieren.
//

inherit "std/thing";

#include <properties.h>
#include <defines.h>
#include <living.h>
#include <wizlevels.h>
#include <food.h>

private static int check_food(int strength);
private static int check_drink(int strength, int alcohol);

int rc;

void reset()
{
    object	env;

    ::reset();
    rc--;
    env=environment(this_object());
    if(env && (rc <= 0))
    if(interactive(env) || env->id("Raum"))
    {  call_out("remove", 0); }
    /* nach 1-3 Resets zerfaellt das Essenobjekt, da sonst echt zuviel
     * Zeugs gehortet werden kann bzw es zulange herumliegt. Essen in NPCs
     * loesen sich nicht auf.
     *
     * das sollte dazu animieren, das Essen gleich einzusetzen und nicht
     * lange zu bunkern. - Samtpfote@Tamedhon.
     */
}

varargs create(commands)
{
  if (IS_BLUE(ME)) return;
  ::create();
  rc=random(3)+1;
  AddId("Nahrung");
  SetProp(P_SHORT, "Ein essbares Objekt");
  SetProp(P_LONG, "Ein essbares Objekt.\n");
  SetProp(P_NAME, "Essen");
  SetProp(P_GENDER, NEUTER);
  SetProp(P_WEIGHT, 50);
  SetProp(P_VALUE, 50);
  SetProp(P_MATERIAL, MAT_FOOD);
  SetProp(P_NOBUY, 1);
  
  if (pointerp(commands) || stringp(commands))
    AddCmd(commands, "mampf");
  else
    AddCmd(({"iss", "esse"}), "mampf");
}

void BeimEssen() {}

int second_life(object player) {}

int mampf(string what)
{
  mapping foodinfo;
  mixed   heal;
  int     hp, sp;

  foodinfo = QueryProp(P_FOOD_INFO);
  notify_fail("Was moechtest Du "+(foodinfo[F_VERB] || "essen")+"?\n");
  if (!what) return 0;
  if (id(what))
  {
    // P_FOOD_INFO auswerten:
    if (pointerp(heal=foodinfo[F_HEAL]) && sizeof(heal)==2)
    { hp = heal[0]; sp = heal[1]; }
    else
      hp = sp = heal;
    if (foodinfo[F_SOAK] && !check_food(foodinfo[F_SOAK]))
      return 1;
    if (foodinfo[F_DRINK] || foodinfo[F_ALCOHOL])
      if (!check_drink(foodinfo[F_DRINK], foodinfo[F_ALCOHOL]))
          return 1;

    if (foodinfo[F_MSG])
      write( break_string(foodinfo[F_MSG], 78, 0, 1) );
    if (foodinfo[F_MSG_ROOM])
      say( break_string(PL->Name(WER)+" "+
          foodinfo[F_MSG_ROOM], 78, 0, 1) );
    BeimEssen();

    if (hp) PL->SetProp(P_HP, PL->QueryProp(P_HP)+hp);
    if (sp) PL->SetProp(P_SP, PL->QueryProp(P_SP)+sp);
    if (foodinfo[F_POISON])
      PL->AddPoison(foodinfo[F_POISON]);

    if (!ME->second_life(PL))
      remove();
    return 1;
  }
  return 0;
}

private static int check_food(int strength)
{
  int food;

  food = PL->QueryProp(P_FOOD) + strength;
  if ((food > PL->QueryProp(P_MAX_FOOD)) && !IS_LEARNER(PL))
  {
    tell_object(PL, "Uff, das bekommst Du nicht mehr 'runter.\n");
    return 0;
  }
  PL->SetProp(P_FOOD, (food < 0 ? 0 : food));
  return 1; 
}

private static int check_drink(int strength, int alcohol)
{
  int drink, alc;

  drink = PL->QueryProp(P_DRINK) + strength;
  alc = PL->QueryProp(P_ALCOHOL) + alcohol;

  if (!IS_LEARNER(PL))
  {
    if (alc > PL->QueryProp(P_MAX_ALCOHOL))
    {
      tell_object(PL,
          "Asso das kriegst Du jezz nich... nichmehr runter, hick.\n");
      return 0;
    }
    if (drink > PL->QueryProp(P_MAX_DRINK))
    {
      tell_object(PL,
          "Soviel kannst Du beim besten Willen nicht mehr trinken...\n");
      return 0;
    }
  }
  PL->SetProp(P_DRINK, (drink < 0 ? 0 : drink));
  PL->SetProp(P_ALCOHOL, (alc < 0 ? 0 : alc));
  return 1; 
}
